@@ -2,10 +2,6 @@
 Will use special scheduler running on rank 0 to orchestrate
 the needed behavior """
 import logging
-
-logging.basicConfig(
-    format="%(asctime)s : %(levelname)s : %(message)s", level=logging.INFO
-)
 from typing import List, Any, Callable
 from mpi4py import MPI
 from functools import partial
@@ -21,76 +17,93 @@ import dill
 comm = MPI.COMM_WORLD
 rank = comm.Get_rank()
 size = comm.Get_size()
-host= MPI.Get_processor_name()
-
-global loop
-loop = True
+host = MPI.Get_processor_name()
 
 MASTER = rank == 0
 
-logging.debug(f"MY {host} RANK {rank} and size {size}")
+logging.debug(f"ME: {host} RANK {rank} and procs {size}")
 
+class begin: 
+    
+    def __init__(self, *args, **kwargs): 
+        logging.debug("[%s][%s] Context init %s",host,rank,kwargs)
+        self.kwargs = kwargs
+            
+    def __enter__(self, *args, **kwargs): 
+        logging.debug("[%s][%s] Context enter",host,rank)
+        try:
+            yield comm
+        finally:
+            pass
+            
+
+    def __exit__(self, exc_type, exc_value, exc_traceback): 
+        logging.debug("[%s][%s] Context exiting %s",host,rank,self.kwargs)
 
-@contextmanager
-def begin(*args, **kwds):
-    try:
-        logging.debug("Yielding comm")
-        yield comm
-    finally:
-        logging.debug("Invoking stop(%s)",rank)
         if rank == 0:
-            logging.debug("kwds %s", kwds)
-            if 'stop' in kwds and kwds['stop']:
-                stop()
-            elif 'stop' not in kwds:
-                stop()
-
+            logging.debug("[%s][%s] Master STOPPING",host,rank)
+            stop()
 
 def mprint(*args):
+    """ Print output if on master node """
     if rank == 0:
         print(*args)
 
 
-def stop():
-    """ Stop all workers """
-    global loop
+def stop(barrier=True):
+    """ Stop all workers from master """
 
     logging.debug("Stopping %s, %s", rank,host)
     if rank == 0:
         logging.debug("Sending break to all ranks")
+        # This will cause each worker rank to hit the barrier
         for i in range(1, size):
-           comm.send("break", dest=i)
-        logging.debug("Waiting on barrier")
-        comm.Barrier()
-        logging.debug("Barrier complete")
+            logging.debug(f"Master sending break to rank {i}")
+            comm.send("break", tag=0, dest=i)
 
+        if barrier:
+            logging.debug("Master Waiting on barrier")
+            comm.Barrier() # Master should barrier until all the workers being stopped barrier too
+            logging.debug("Barrier complete")
+
+        logging.debug("Master STOP complete!")
 
 if rank != 0:
+    """ Monitor thread for (worker processes) receiving function tasks to execute """
     def run():
-        while loop:
-            logging.debug("thread rank %s waiting on defer", rank)
-            defer = comm.recv(source=0)
-            logging.debug("thread rank %s got data %s", rank, defer)
+        while True:
+            logging.debug("[%s] thread rank %s waiting on defer",host,  rank)
+            logging.debug("[%s][%s] RECV BEFORE TAG=0",host,rank)
+
+            # Listen for tasks and "break" commands from Master
+            defer = comm.recv(source=0, tag=0)
+
+            logging.debug("[%s][%s] RECV AFTER TAG=0",host,rank)
+            logging.debug("[%s] thread rank %s got defer",host,  rank)
+            logging.debug("[%s] thread rank %s got data %s", host, rank, defer)
+
             if type(defer) is str and defer == "break":
-                logging.debug("Rank %s stopping", rank)
+                logging.debug("[%s] thread rank %s breaking", host, rank)
                 break
+
             defer = dill.loads(defer)
-            logging.debug("thread rank %s got defer", defer)
+            logging.debug("[%s] thread rank %s got defer",host,  defer)
             result = defer()
-            logging.debug("thread rank %s sending result %s", rank, result)
-            comm.send(result, dest=0)
-        logging.debug(f"Rank {rank} notifying Barrier")
+            logging.debug("[%s] thread rank %s sending result %s",host,  rank, result)
+            comm.send(result, tag=0, dest=0)
+
+        logging.debug(f"{host} Rank {rank} notifying Barrier")
         comm.Barrier()
 
+        logging.debug("[%s] thread rank [%s] message loop ending", host, rank)
 
     thread = Thread(target=run)
     thread.start()
 
 
 def parallel(defers: List, *args):
-    """ This will use the master node 0 scheduler to scatter/gather results """
-    _rank = 1
-    logging.debug("parallel rank %s %s", rank, args)
+    """ Run list of tasks in parallel across compute fabric """
+    logging.debug("[%s] parallel rank %s %s",host, rank, args)
     l = len(defers)
 
     if rank == 0:
@@ -104,10 +117,10 @@ def parallel(defers: List, *args):
         # Send all functions to other ranks to execute in parallel
         for defer in defers:
             if last_result is None:
-                logging.debug("parallel Sending defer %s to rank %s NO ARGS", defer, dest_rank)
-                comm.send(dill.dumps(defer), dest=dest_rank)
+                logging.debug("[%s] parallel Sending defer %s to rank %s NO ARGS",host,  defer, dest_rank)
+                comm.send(dill.dumps(defer), tag=0, dest=dest_rank)
             else:
-                logging.debug("parallel Sending defer %s to rank %s %s", defer, dest_rank, last_result)
+                logging.debug("[%s] parallel Sending defer %s to rank %s %s",host,  defer, dest_rank, last_result)
                 fname = None
                 if type(defer) is partial:
                     fname = defer.func.__name__
@@ -125,14 +138,14 @@ def parallel(defers: List, *args):
 
                 logging.debug("FNAME %s", fname)
                 if fname in ['parallel', 'pipeline']:
-                    logging.debug("PARALLEL EXECUTING %s %s", defer, last_result)
+                    logging.debug("[%s] PARALLEL EXECUTING %s %s",host,  defer, last_result)
                     # Just execute the parallel since I am already at rank 0
                     last_result = defer(*last_result)
                     l = l - 1
-                    logging.debug("PARALLEL %s EXECUTING DONE %s", fname, last_result)
+                    logging.debug("[%s] PARALLEL %s EXECUTING DONE %s",host,  fname, last_result)
                     continue
                 else:
-                    comm.send(dill.dumps(partial(defer, *last_result)), dest=dest_rank)
+                    comm.send(dill.dumps(partial(defer, *last_result)), tag=0, dest=dest_rank)
 
             # Cycle destination rank over available nodes
             dest_rank += 1
@@ -142,16 +155,19 @@ def parallel(defers: List, *args):
         # After all functions are sent and being executed, get results over MPI comm
         results = []
         for i in range(0, l):
-            logging.debug("parallel Master Waiting on result")
-            last_result = comm.recv()
-            logging.debug("parallel Master Got result %s", last_result)
+            logging.debug("[%s] parallel Master Waiting on result", host)
+            logging.debug("[%s][%s] RECV BEFORE TAG=0",host,rank)
+            last_result = comm.recv(tag=0)
+            logging.debug("[%s][%s] RECV AFTER TAG=0",host,rank)
+            logging.debug("[%s] parallel Master Got result %s", host, last_result)
             results += [last_result]
 
-        logging.debug("parallel return results %s", results)
+        logging.debug("[%s] parallel return results %s", host, results)
         return results
 
 
 def enumrate(gen):
+    """ Alternate enumerate as a generator """
     i = 0
     for a in gen:
         yield i, a
@@ -167,22 +183,25 @@ def mapreduce(_map: Callable, _reduce: Callable, data: Any, require_list=False):
         _chunks = np.array_split(data, size)
     else:
         _chunks = None
+
     data = comm.scatter(_chunks, root=0)
 
     listdata = data.tolist()
-    logging.info("MAP: %s",listdata)
+    logging.debug("MAP: %s",listdata)
     _data = _map(listdata)
     newData = comm.gather(_data, root=0)
     results += [newData]
     _flattened = flatten(results)
     if None not in _flattened:
-        logging.info("REDUCE: %s",_flattened)
+        logging.debug("[%s] REDUCE: %s",host, _flattened)
         _data = _reduce(_flattened)
         if require_list and type(_data) is list:
             mapreduce(_map, _reduce, _data)
         return _data
 
 def map(func: Callable, data: Any):
+    """ Apply map function over data elements """
+    """ Runs in parallel over data """
     _funcs = []
     for arg in data:
         if iterable(arg):
@@ -197,6 +216,9 @@ def map(func: Callable, data: Any):
 
 
 def reduce(func: Callable, data: Any):
+    """ Apply reduce function over data elements """
+    """ Results cascade to next reduce function """
+    """ Runs in sequence over data """
     _funcs = []
     if data is None:
         return None
@@ -237,14 +259,14 @@ def scatter(data: Any, func: Callable):
 
         data = comm.scatter(chunk, root=0)
         _data = func(data)
-        logging.debug("scatter[%s, %s]: Chunk %s %s, Func is %s Data is %s Result is %s", rank,host,i, chunk, func, data, _data)
+        logging.debug("[%s] scatter[%s, %s]: Chunk %s %s, Func is %s Data is %s Result is %s", host, rank,host,i, chunk, func, data, _data)
         newData = comm.gather(_data, root=0)
         results += [newData]
 
     return flatten(results)
 
 def pipeline(defers: List, *args):
-    """ This will use the master node 0 scheduler to orchestrate results """
+    """ Run list of functions in ordered sequence, passing intermediate results on to next task """
     logging.debug("pipeline rank %s %s", rank, args)
     if rank == 0:
         dest_rank = 1
@@ -256,8 +278,8 @@ def pipeline(defers: List, *args):
 
         for defer in defers:
             if last_result is None:
-                logging.debug("Pipeline Sending defer to rank %s NO ARGS", dest_rank)
-                comm.send(dill.dumps(defer), dest=dest_rank)
+                logging.debug("[%s] Pipeline Sending defer to rank %s NO ARGS", host, dest_rank)
+                comm.send(dill.dumps(defer), tag=0, dest=dest_rank)
             else:
                 logging.debug("Pipeline Sending defer %s to rank %s %s", defer, dest_rank, last_result)
                 fname = None
@@ -280,7 +302,7 @@ def pipeline(defers: List, *args):
                     last_result = defer(last_result)
                     continue
                 else:
-                    comm.send(dill.dumps(partial(defer, last_result)), dest=dest_rank)
+                    comm.send(dill.dumps(partial(defer, last_result)), tag=0, dest=dest_rank)
 
             dest_rank += 1
 
@@ -288,7 +310,7 @@ def pipeline(defers: List, *args):
                 dest_rank = 1
 
             logging.debug("Pipeline Master Waiting on result")
-            last_result = comm.recv()
+            last_result = comm.recv(tag=0)
             logging.debug("Pipeline Master Got result %s", last_result)
 
         return last_result
